--  Generic logical, relational and arithmetic operations that can
--  handle the value 'unknown'.
--  If the operands are numbers, the same results as from standard
--  Euphoria operations will be returnd.

--  v1.0 (Euphoria 2.3), 25. November 2002
--  by Juergen Luethje <support {AT} luethje {DOT} eu>
--  Standard disclaimer: Use at your own risk!

--  Thanks:
--  o Derek Parnell, and Pete Lomax for helping me, especially with
--    the truth table.
--  o Robert Craig for the powerful elegance of Euphoria.


--  === Contents of this file ===
--
--  * Constants *
--  FALSE
--  TRUE
--  UNKNOWN
--
--  * Type *
--  uNum         -- numeric value or UNKNOWN
--
--  * Logical operations *
--  u_not(a)
--  u_and(a, b)
--  u_or (a, b)
--  u_xor(a, b)
--
--  * Relational Operations *
--  u_equ(a, b)  -- a equal to b
--  u_lt (a, b)  -- a less than b
--  u_lte(a, b)  -- a less than or equal to b
--  u_gt (a, b)  -- a greater than b
--  u_gte(a, b)  -- a greater than or equal to b
--
--  * Arithmetic Operations *
--  u_neg(a)     -- -a
--  u_add(a, b)  -- a + b
--  u_sub(a, b)  -- a - b
--  u_mul(a, b)  -- a * b
--  u_div(a, b)  -- a / b   Note: u_div(a, 0) = UNKNOWN
--
--
--  === Remarks ===
--
--  Truth table for three-valued logic
--  (T = True, F = False, U = Unknown):
--
--  A | B | not A | A and B | A or B | A xor B
--  --+---+-------+---------+--------+--------
--  T | T |   F   |    T    |   T    |    F
--  T | F |   F   |    F    |   T    |    T
--  T | U |   F   |    U    |   T    |    U
--  F | T |   T   |    F    |   T    |    T
--  F | F |   T   |    F    |   F    |    F
--  F | U |   T   |    F    |   U    |    U
--  U | T |   U   |    U    |   T    |    U
--  U | F |   U   |    F    |   U    |    U
--  U | U |   U   |    U    |   U    |    U
--
--
--  Relational and arithmetic operations will always produce UNKNOWN
--  when at least one of the operands is UNKNOWN.


--====================================================================--

global constant
   FALSE   =  0,
   TRUE    = not FALSE,
   UNKNOWN = {}


global type uNum (object x)
   -- numeric value or UNKNOWN
   if atom(x) or equal(x, UNKNOWN) then
      return TRUE
   end if
   return FALSE
end type


-------------------------[ Logical operations ]-------------------------

global function u_not (uNum a)
   if equal(a, UNKNOWN) then
      return UNKNOWN
   else
      return not a
   end if
end function


global function u_and (uNum a, uNum b)
   if equal(a, FALSE) or equal(b, FALSE) then
      return FALSE
   elsif equal(a, UNKNOWN) or equal(b, UNKNOWN) then
      return UNKNOWN
   else
      return TRUE
   end if
end function


global function u_or (uNum a, uNum b)
   if equal(not a, FALSE) or equal(not b, FALSE) then
      return TRUE
   elsif equal(a, UNKNOWN) or equal(b, UNKNOWN) then
      return UNKNOWN
   else
      return FALSE
   end if
end function


global function u_xor (uNum a, uNum b)
   if equal(a, UNKNOWN) or equal(b, UNKNOWN) then
      return UNKNOWN
   else
      return a xor b
   end if
end function


-----------------------[ Relational Operations ]------------------------

global function u_equ (uNum a, uNum b)
   -- a equal to b
   if equal(a, UNKNOWN) or equal(b, UNKNOWN) then
      return UNKNOWN
   else
      return a = b
   end if
end function


global function u_lt (uNum a, uNum b)
   -- a less than b
   if equal(a, UNKNOWN) or equal(b, UNKNOWN) then
      return UNKNOWN
   else
      return a < b
   end if
end function


global function u_lte (uNum a, uNum b)
   -- a less than or equal to b
   if equal(a, UNKNOWN) or equal(b, UNKNOWN) then
      return UNKNOWN
   else
      return a <= b
   end if
end function


global function u_gt (uNum a, uNum b)
   -- a a greater than b
   if equal(a, UNKNOWN) or equal(b, UNKNOWN) then
      return UNKNOWN
   else
      return a > b
   end if
end function


global function u_gte (uNum a, uNum b)
   -- a greater than or equal to b
   if equal(a, UNKNOWN) or equal(b, UNKNOWN) then
      return UNKNOWN
   else
      return a >= b
   end if
end function


-----------------------[ Arithmetic Operations ]------------------------

global function u_neg (uNum a)
   -- Negation
   if equal(a, UNKNOWN) then
      return UNKNOWN
   else
      return -a
   end if
end function


global function u_add (uNum a, uNum b)
   -- Addition
   if equal(a, UNKNOWN) or equal(b, UNKNOWN) then
      return UNKNOWN
   else
      return a + b
   end if
end function


global function u_sub (uNum a, uNum b)
   -- Subtraction
   if equal(a, UNKNOWN) or equal(b, UNKNOWN) then
      return UNKNOWN
   else
      return a - b
   end if
end function


global function u_mul (uNum a, uNum b)
   -- Multiplication
   if equal(a, UNKNOWN) or equal(b, UNKNOWN) then
      return UNKNOWN
   else
      return a * b
   end if
end function


global function u_div (uNum a, uNum b)
   -- Division
   if equal(a, UNKNOWN) or equal(b, UNKNOWN) or b = 0 then
      return UNKNOWN
   else
      return a / b
   end if
end function
